@@ -1,7 +1,8 @@
 """Built-in commands for Slashed."""
 
-from slashed.base import BaseCommand
-from slashed.builtin.help_cmd import help_cmd, exit_cmd
+from __future__ import annotations
+
+from slashed.builtin.help_cmd import HelpCommand, ExitCommand
 from slashed.builtin.system import (
     ExecCommand,
     ProcessesCommand,
@@ -10,11 +11,15 @@ from slashed.builtin.system import (
     KillCommand,
     EnvCommand,
 )
+from typing import TYPE_CHECKING
+
+if TYPE_CHECKING:
+    from slashed.base import BaseCommand
 
 
 def get_builtin_commands() -> list[BaseCommand]:
     """Get list of built-in commands."""
-    return [help_cmd, exit_cmd]
+    return [HelpCommand(), ExitCommand()]
 
 
 def get_system_commands() -> list[BaseCommand]:
