@@ -2,15 +2,13 @@
 
 from __future__ import annotations
 
+from collections.abc import Awaitable, Iterable
 from copy import copy
+import inspect
 import os
-from typing import TYPE_CHECKING, Any
+from typing import TYPE_CHECKING, Any, TypeGuard, cast
 
-
-try:
-    from upath import UPath as Path
-except ImportError:
-    from pathlib import Path
+from upath import UPath
 
 from slashed import utils
 from slashed.completion import CompletionItem, CompletionProvider
@@ -21,7 +19,14 @@ type PathType = str | os.PathLike[str]
 
 
 if TYPE_CHECKING:
-    from collections.abc import Callable, Iterable, Iterator, Sequence
+    from collections.abc import (
+        AsyncIterator,
+        Awaitable,
+        Callable,
+        Sequence,
+    )
+
+    from upath.types import JoinablePathLike
 
     from slashed.completion import CompletionContext
     from slashed.slashed_types import CompletionKind
@@ -29,6 +34,11 @@ if TYPE_CHECKING:
 logger = get_logger(__name__)
 
 
+def is_awaitable[T](obj: Awaitable[T] | T) -> TypeGuard[Awaitable[T]]:
+    """Type guard to check if object is awaitable."""
+    return inspect.isawaitable(obj)
+
+
 class PathCompleter(CompletionProvider):
     """Provides filesystem path completions."""
 
@@ -40,7 +50,7 @@ class PathCompleter(CompletionProvider):
         files: bool = True,
         show_hidden: bool = False,
         expanduser: bool = True,
-        base_path: str | os.PathLike[str] | None = None,
+        base_path: JoinablePathLike | None = None,
     ):
         """Initialize path completer.
 
@@ -57,26 +67,26 @@ class PathCompleter(CompletionProvider):
         self.files = files
         self.show_hidden = show_hidden
         self.expanduser = expanduser
-        self.base_path = Path(base_path).resolve() if base_path else None
+        self.base_path = UPath(base_path).resolve() if base_path else None
 
-    def get_completions(
+    async def get_completions(
         self,
         context: CompletionContext,
-    ) -> Iterator[CompletionItem]:
+    ) -> AsyncIterator[CompletionItem]:
         """Get path completions."""
         word = context.current_word or "."
 
         try:
             # Handle absolute paths
-            if Path(word).is_absolute():
-                path = Path(word)
+            if UPath(word).is_absolute():
+                path = UPath(word)
             # Handle user paths
             elif self.expanduser and word.startswith("~"):
-                path = Path(word).expanduser()
+                path = UPath(word).expanduser()
             # Handle relative paths
             else:
                 # If base_path is set, resolve relative to it
-                base = self.base_path or Path.cwd()
+                base = self.base_path or UPath.cwd()
                 path = (base / word).resolve()
 
             # If path doesn't exist, use its parent for listing
@@ -112,7 +122,7 @@ class PathCompleter(CompletionProvider):
                 # Get relative path if using base_path
                 if self.base_path:
                     try:
-                        display = str(entry.relative_to(self.base_path))
+                        display = str(entry.relative_to(self.base_path))  # pyright: ignore[reportArgumentType]
                     except ValueError:
                         display = str(entry)
                 else:
@@ -150,10 +160,10 @@ class EnvVarCompleter(CompletionProvider):
         self.prefixes = prefixes
         self.include_values = include_values
 
-    def get_completions(
+    async def get_completions(
         self,
         context: CompletionContext,
-    ) -> Iterator[CompletionItem]:
+    ) -> AsyncIterator[CompletionItem]:
         """Get environment variable completions."""
         word = context.current_word.lstrip("$")
 
@@ -188,10 +198,10 @@ class ChoiceCompleter(CompletionProvider):
             self.choices = list(choices)
             self.descriptions = {}
 
-    def get_completions(
+    async def get_completions(
         self,
         context: CompletionContext,
-    ) -> Iterator[CompletionItem]:
+    ) -> AsyncIterator[CompletionItem]:
         """Get matching choices."""
         word = context.current_word
         if self.ignore_case:
@@ -225,28 +235,37 @@ class MultiValueCompleter(CompletionProvider):
         self.delimiter = delimiter
         self.strip = strip
 
-    def get_completions(
+    async def get_completions(
         self,
         context: CompletionContext,
-    ) -> Iterator[CompletionItem]:
+    ) -> AsyncIterator[CompletionItem]:
         """Get completions for current value."""
-        # Split into values and get current value
-        values = context.current_word.split(self.delimiter)
-        current = values[-1]
+        # Use full text instead of just current_word for splitting
+        full_text = context._document.text
+        values = full_text.split(self.delimiter)
+
+        # Current value is what's being typed
+        current = context.current_word
         if self.strip:
             current = current.strip()
+
+        # Create modified context with just the current value
         mod_context = copy(context)
         mod_context._current_word = current  # type: ignore[attr-defined]
 
-        # Get completions from base provider
-        prefix = self.delimiter.join(values[:-1])
-        if prefix:
+        # Build prefix from previous values if we have more than one value
+        prefix = ""
+        if len(values) > 1:
+            prefix = self.delimiter.join(values[:-1])
             prefix = f"{prefix}{self.delimiter}"
             if self.strip:
                 prefix = f"{prefix} "
 
-        for item in self.provider.get_completions(mod_context):
-            item.text = f"{prefix}{item.text}"
+        # Get completions and add prefix
+        async for item in self.provider.get_completions(mod_context):
+            item = copy(item)  # Don't modify original item
+            if prefix:
+                item.text = f"{prefix}{item.text}"
             yield item
 
 
@@ -267,10 +286,10 @@ class KeywordCompleter(CompletionProvider):
         self.keywords = keywords
         self.value_provider = value_provider
 
-    def get_completions(
+    async def get_completions(
         self,
         context: CompletionContext,
-    ) -> Iterator[CompletionItem]:
+    ) -> AsyncIterator[CompletionItem]:
         """Get keyword completions."""
         word = context.current_word
 
@@ -286,7 +305,8 @@ class KeywordCompleter(CompletionProvider):
         if self.value_provider and context.arg_position > 0:
             prev_arg = context.command_args[context.arg_position - 1]
             if prev_arg.startswith("--"):
-                yield from self.value_provider.get_completions(context)
+                async for item in self.value_provider.get_completions(context):
+                    yield item
 
 
 class ChainedCompleter(CompletionProvider):
@@ -300,13 +320,14 @@ class ChainedCompleter(CompletionProvider):
         """
         self.providers = providers
 
-    def get_completions(
+    async def get_completions(
         self,
         context: CompletionContext,
-    ) -> Iterator[CompletionItem]:
+    ) -> AsyncIterator[CompletionItem]:
         """Get completions from all providers."""
         for provider in self.providers:
-            yield from provider.get_completions(context)
+            async for item in provider.get_completions(context):
+                yield item
 
 
 class CallbackCompleter(CompletionProvider):
@@ -328,25 +349,33 @@ class CallbackCompleter(CompletionProvider):
 
     def __init__(
         self,
-        callback: Callable[[CompletionContext], Iterable[str | CompletionItem]],
+        callback: (
+            Callable[[CompletionContext], Iterable[str | CompletionItem]]
+            | Callable[[CompletionContext], Awaitable[Iterable[str | CompletionItem]]]
+        ),
         kind: CompletionKind | None = None,
     ):
-        """Initialize callback completer.
+        self._callback = callback
+        self._kind = kind
+        self._is_async = inspect.iscoroutinefunction(callback)
 
-        Args:
-            callback: Function that returns completions as strings or CompletionItems
-            kind: Optional kind to apply to string completions
-        """
-        self.callback = callback
-        self.kind = kind
-
-    def get_completions(
+    async def get_completions(
         self,
         context: CompletionContext,
-    ) -> Iterator[CompletionItem]:
-        """Get completions by calling the callback function."""
-        for item in self.callback(context):
-            if isinstance(item, str):
-                yield CompletionItem(text=item, kind=self.kind)  # pyright: ignore
+    ) -> AsyncIterator[CompletionItem]:
+        try:
+            result = self._callback(context)
+
+            # Narrow type and handle async/sync appropriately
+            if is_awaitable(result):
+                items = await result
             else:
-                yield item
+                items = cast(Iterable[str | CompletionItem], result)
+
+            for item in items:
+                if isinstance(item, str):
+                    yield CompletionItem(text=item, kind=self._kind)  # type: ignore
+                else:
+                    yield item
+        except Exception as e:  # noqa: BLE001
+            logger.debug("Completion callback error: %s", e)
