@@ -5,33 +5,27 @@ from __future__ import annotations
 from typing import TYPE_CHECKING, Any
 
 from prompt_toolkit.completion import Completer, Completion
-from typing_extensions import TypeVar
 
-from slashed import CommandContext, CompletionContext
+from slashed import CompletionContext
 from slashed.log import get_logger
 from slashed.store import CommandStore
 from slashed.utils import get_first_line
 
 
 if TYPE_CHECKING:
-    from collections.abc import Iterator
+    from collections.abc import AsyncGenerator, Iterable
 
     from prompt_toolkit.document import Document
 
+    from slashed import CommandContext
     from slashed.base import OutputWriter
 
 
-TContextData = TypeVar("TContextData", default=Any)
 logger = get_logger(__name__)
 
 
-class PromptToolkitCompleter[TContextData](Completer):
-    """Adapts our completion system to prompt-toolkit.
-
-    Type Parameters:
-        TContextData: Type of the data in the associated CommandContext. Used when
-                     completions need to access typed context data.
-    """
+class PromptToolkitCompleter[TContextData = Any](Completer):
+    """Adapts our completion system to prompt-toolkit."""
 
     def __init__(
         self,
@@ -55,10 +49,14 @@ class PromptToolkitCompleter[TContextData](Completer):
         )
 
     def get_completions(
-        self,
-        document: Document,
-        complete_event: Any,
-    ) -> Iterator[Completion]:
+        self, document: Document, complete_event: Any
+    ) -> Iterable[Completion]:
+        """Sync completions - not used but required by interface."""
+        return []
+
+    async def get_completions_async(
+        self, document: Document, complete_event: Any
+    ) -> AsyncGenerator[Completion]:
         """Get completions for the current context."""
         text = document.text.lstrip()
 
@@ -77,13 +75,13 @@ class PromptToolkitCompleter[TContextData](Completer):
                     completer := command.get_completer()
                 ):
                     current_word = completion_context.current_word
-                    for item in completer.get_completions(completion_context):
+                    async for item in completer.get_completions(completion_context):
                         if current_word and not item.text.startswith(current_word):
                             continue
                         start_pos = -len(current_word) if current_word else 0
                         yield Completion(
                             item.text,
-                            start_pos,
+                            start_position=start_pos,
                             display_meta=get_first_line(item.metadata),
                         )
                     return
@@ -93,7 +91,9 @@ class PromptToolkitCompleter[TContextData](Completer):
             for cmd in self._store.list_commands():
                 if cmd.name.startswith(word):
                     yield Completion(
-                        cmd.name, -len(word), display_meta=get_first_line(cmd.description)
+                        cmd.name,
+                        start_position=-len(word),
+                        display_meta=get_first_line(cmd.description),
                     )
 
         except RuntimeError as e:
