@@ -8,29 +8,24 @@ from typing import TYPE_CHECKING, Any, TypeVar
 
 from psygnal import Signal
 from psygnal.containers import EventedDict
+from upath import UPath
 
 from slashed.base import (
     BaseCommand,
     Command,
     CommandContext,
     ExecuteFunc,
-    OutputWriter,
     parse_command,
 )
 from slashed.builtin import get_system_commands
-from slashed.completion import CompletionContext, CompletionProvider
+from slashed.completion import CompletionContext
 from slashed.context import ContextRegistry
 from slashed.events import CommandExecutedEvent
 from slashed.exceptions import CommandError
 from slashed.log import get_logger
-from slashed.output import DefaultOutputWriter, SignalingOutputWriter
+from slashed.output import CallbackOutputWriter, DefaultOutputWriter
 
 
-try:
-    from upath import UPath as Path
-except ImportError:
-    from pathlib import Path
-
 if TYPE_CHECKING:
     from collections.abc import Callable
     import os
@@ -38,10 +33,11 @@ if TYPE_CHECKING:
     from prompt_toolkit.document import Document
     from psygnal.containers._evented_dict import DictEvents
 
+    from slashed.base import OutputWriter
     from slashed.commands import SlashedCommand
+    from slashed.completion import CompletionProvider
 
 
-TContextData = TypeVar("TContextData")
 logger = get_logger(__name__)
 TCommandFunc = TypeVar("TCommandFunc", bound=ExecuteFunc)
 
@@ -68,7 +64,7 @@ class CommandStore:
         self._commands = EventedDict[str, BaseCommand]()
         self._contexts = ContextRegistry()
         self._command_history: list[str] = []
-        self._history_path = Path(history_file) if history_file else None
+        self._history_path = UPath(history_file) if history_file else None
         self._enable_system_commands = enable_system_commands
         self._initialized = False
         if self._history_path:
@@ -90,7 +86,7 @@ class CommandStore:
         # Load history
         try:
             if self._history_path and self._history_path.exists():
-                self._command_history = self._history_path.read_text().splitlines()
+                self._command_history = self._history_path.read_text("utf-8").splitlines()
         except Exception:
             logger.exception("Failed to load command history")
             self._command_history = []
@@ -124,7 +120,7 @@ class CommandStore:
     def create_context[TContextData](
         self,
         data: TContextData | None,
-        output_writer: OutputWriter | None = None,
+        output_writer: OutputWriter | Callable[..., Any] | None = None,
         metadata: dict[str, Any] | None = None,
     ) -> CommandContext[TContextData]:
         """Create a command execution context.
@@ -137,12 +133,14 @@ class CommandStore:
         Returns:
             Command execution context
         """
-        base_writer = output_writer or DefaultOutputWriter()
-        writer = SignalingOutputWriter(self.output, base_writer)
+        if callable(output_writer):
+            writer: OutputWriter = CallbackOutputWriter(output_writer)
+        else:
+            writer = output_writer or DefaultOutputWriter()
         meta = metadata or {}
         return CommandContext(output=writer, data=data, command_store=self, metadata=meta)
 
-    def create_completion_context(
+    def create_completion_context[TContextData](
         self,
         document: Document,
         command_context: CommandContext[TContextData] | None = None,
@@ -244,10 +242,9 @@ class CommandStore:
         if not command.is_available():
             return
         if command.name in self._commands:
-            msg = f"Command '{command.name}' already registered"
+            msg = f"Command {command.name!r} already registered"
             raise ValueError(msg)
         self._commands[command.name] = command
-        logger.debug("Registered command: %s", command.name)
 
     def unregister_command(self, name: str):
         """Remove a command.
@@ -302,7 +299,11 @@ class CommandStore:
             result.setdefault(cmd.category, []).append(cmd)
         return result
 
-    async def execute_command(self, command_str: str, ctx: CommandContext[TContextData]):
+    async def execute_command[TContextData](
+        self,
+        command_str: str,
+        ctx: CommandContext[TContextData],
+    ):
         """Execute a command from string input.
 
         Args:
@@ -349,7 +350,7 @@ class CommandStore:
                         ])
                     if command.help_text:
                         sections.extend(["Help:", command.help_text])
-                    await ctx.output.print("\n".join(sections))
+                    await ctx.print("\n".join(sections))
                     return
 
             msg = "Executing command: %s (args=%s, kwargs=%s)"
@@ -379,7 +380,7 @@ class CommandStore:
         self,
         command_str: str,
         context: T | None = None,  # type: ignore[type-var]
-        output_writer: OutputWriter | None = None,
+        output_writer: OutputWriter | Callable[..., Any] | None = None,
         metadata: dict[str, Any] | None = None,
     ):
         """Execute a command with a custom context.
@@ -503,7 +504,7 @@ class CommandStore:
             @store.command(category="tools")
             async def hello(ctx: CommandContext, name: str = "World"):
                 '''Say hello to someone.'''
-                await ctx.output.print(f"Hello {name}!")
+                await ctx.print(f"Hello {name}!")
             ```
         """
 
